
/* Chrysalide - Outil d'analyse de fichiers binaires
 * params.c - équivalent Python du fichier "core/params.c"
 *
 * Copyright (C) 2014-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "params.h"


#include <pygobject.h>


#include <core/params.h>


#include "../helpers.h"



/* Fournit la version du programme global. */
static PyObject *py_params_get_main_configuration(PyObject *, PyObject *);

/* Définit les constantes pour les paramètres. */
static bool py_params_define_constants(PyTypeObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : self = NULL car méthode statique.                            *
*                args = non utilisé ici.                                      *
*                                                                             *
*  Description : Fournit la version du programme global.                      *
*                                                                             *
*  Retour      : Numéro de révision.                                          *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_params_get_main_configuration(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Instance GLib à retourner   */
    GGenConfig *config;                     /* Configuration à convertir   */

    config = get_main_configuration();

    result = pygobject_new(G_OBJECT(config));
    Py_XINCREF(result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_params_type(void)
{
    static PyMethodDef py_params_methods[] = {

        { "get_main_configuration", py_params_get_main_configuration,
          METH_NOARGS | METH_STATIC,
          "Give access to the main configuration of Chrysalide."
        },
        { NULL }

    };

    static PyTypeObject py_params_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name = "pychrysalide.core.params",
        .tp_basicsize = sizeof(PyObject),

        .tp_flags = Py_TPFLAGS_DEFAULT,

        .tp_doc = "Python object for parameters",

        .tp_methods =  py_params_methods

    };

    return &py_params_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : obj_type = type dont le dictionnaire est à compléter.        *
*                                                                             *
*  Description : Définit les constantes pour les paramètres.                  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool py_params_define_constants(PyTypeObject *obj_type)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    result &= PyDict_AddStringMacro(obj_type, MPK_LAST_PROJECT);
    result &= PyDict_AddStringMacro(obj_type, MPK_ELLIPSIS_HEADER);
    result &= PyDict_AddStringMacro(obj_type, MPK_ELLIPSIS_TAB);
    result &= PyDict_AddStringMacro(obj_type, MPK_KEYBINDINGS_EDIT);
    result &= PyDict_AddStringMacro(obj_type, MPK_AUTO_SAVE);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.core.params'.          *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_params(PyObject *module)
{
    PyTypeObject *py_params_type;           /* Type Python pour 'params'   */
    int ret;                                /* Bilan d'un appel            */

    py_params_type = get_python_params_type();

    py_params_type->tp_new = PyType_GenericNew;

    if (PyType_Ready(py_params_type) != 0)
        return false;

    if (!py_params_define_constants(py_params_type))
        return false;

    Py_INCREF(py_params_type);
    ret = PyModule_AddObject(module, "params", (PyObject *)py_params_type);

    return (ret == 0);

}
